﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.IO;
using System.Linq;
using System.Net;
using Delay.Web.Helpers;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace AutomatedTests
{
    [TestClass]
    public class AmazonS3StorageTests
    {
        private const string TestBucketName = "amazons3storagetestbucket";
        private const string TestBlobName = "TestBlob.txt";
        private const string TestBlobAddress = TestBucketName + "/" + TestBlobName;
        private const string TestString = "This is a test string\x203c"; // \x203c is Double Exclamation Mark
        private readonly byte[] TestBytes = new byte[10000];

        public AmazonS3StorageTests()
        {
            Random rand = new Random(1);
            rand.NextBytes(TestBytes);
        }

        [TestInitialize]
        public void TestInitialize()
        {
            // TODO: Add your credentials here
            AmazonS3Storage.AccessKeyId = "Your Access Key ID";
            AmazonS3Storage.SecretAccessKey = "Your Secret Access Key";
        }

        [TestCleanup]
        public void TestCleanup()
        {
            AmazonS3Storage.DeleteBlob(TestBlobAddress);
        }

        [TestMethod]
        public void ListCreateDeleteBlobContainer()
        {
            var temporaryContainerName = TestBucketName + "-" + DateTime.UtcNow.Ticks;
            Assert.IsFalse(AmazonS3Storage.ListBlobContainers().Contains(temporaryContainerName));
            AmazonS3Storage.CreateBlobContainer(temporaryContainerName);
            Assert.IsTrue(AmazonS3Storage.ListBlobContainers().Contains(temporaryContainerName));
            AmazonS3Storage.DeleteBlobContainer(temporaryContainerName);
            Assert.IsFalse(AmazonS3Storage.ListBlobContainers().Contains(temporaryContainerName));
        }

        [TestMethod]
        public void CreatedBucketNamesAlwaysLowerCase()
        {
            var temporaryContainerName = TestBucketName + "-" + DateTime.UtcNow.Ticks;
            AmazonS3Storage.CreateBlobContainer(temporaryContainerName.ToUpperInvariant());
            Assert.IsTrue(AmazonS3Storage.ListBlobContainers().Contains(temporaryContainerName));
            AmazonS3Storage.DeleteBlobContainer(temporaryContainerName);
        }

        [TestMethod]
        public void DuplicateBucketCreate()
        {
            AmazonS3Storage.CreateBlobContainer(TestBucketName);
        }

        [TestMethod]
        public void ListBlobsDeleteBlob1()
        {
            var temporaryBlobAddress = TestBlobAddress + "-" + DateTime.UtcNow.Ticks;
            Assert.IsFalse(AmazonS3Storage.ListBlobs(TestBucketName).Contains(temporaryBlobAddress));
            AmazonS3Storage.UploadTextToBlob(temporaryBlobAddress, TestString);
            Assert.IsTrue(AmazonS3Storage.ListBlobs(TestBucketName).Contains(temporaryBlobAddress));
            AmazonS3Storage.DeleteBlob(temporaryBlobAddress);
            Assert.IsFalse(AmazonS3Storage.ListBlobs(TestBucketName).Contains(temporaryBlobAddress));
        }

        [TestMethod]
        public void ListBlobsDeleteBlob2()
        {
            var uniqueSuffix = DateTime.UtcNow.Ticks;
            var temporaryBlobAddress = TestBlobAddress + "-" + uniqueSuffix;
            var temporaryBlobName = TestBlobName + "-" + uniqueSuffix;
            Assert.IsFalse(AmazonS3Storage.ListBlobs(TestBucketName).Contains(temporaryBlobAddress));
            AmazonS3Storage.UploadTextToBlob(TestBucketName, temporaryBlobName, TestString);
            Assert.IsTrue(AmazonS3Storage.ListBlobs(TestBucketName).Contains(temporaryBlobAddress));
            AmazonS3Storage.DeleteBlob(TestBucketName, temporaryBlobName);
            Assert.IsFalse(AmazonS3Storage.ListBlobs(TestBucketName).Contains(temporaryBlobAddress));
        }

        [TestMethod]
        public void UploadDownloadTextToBlob1()
        {
            var uploadString = TestString;
            AmazonS3Storage.UploadTextToBlob(TestBlobAddress, uploadString);
            var downloadString = AmazonS3Storage.DownloadBlobAsText(TestBlobAddress);
            Assert.AreEqual(uploadString, downloadString);
        }

        [TestMethod]
        public void UploadDownloadTextToBlob2()
        {
            var uploadString = TestString;
            AmazonS3Storage.UploadTextToBlob(TestBucketName, TestBlobName, uploadString);
            var downloadString = AmazonS3Storage.DownloadBlobAsText(TestBucketName, TestBlobName);
            Assert.AreEqual(uploadString, downloadString);
        }

        [TestMethod]
        public void UploadDownloadBinaryToBlobBytes1()
        {
            var uploadBytes = TestBytes;
            AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, uploadBytes);
            var downloadBytes = AmazonS3Storage.DownloadBlobAsByteArray(TestBlobAddress);
            Assert.IsTrue(Enumerable.SequenceEqual(uploadBytes, downloadBytes));
        }

        [TestMethod]
        public void UploadDownloadBinaryToBlobBytes2()
        {
            var uploadBytes = TestBytes;
            AmazonS3Storage.UploadBinaryToBlob(TestBucketName, TestBlobName, uploadBytes);
            var downloadBytes = AmazonS3Storage.DownloadBlobAsByteArray(TestBucketName, TestBlobName);
            Assert.IsTrue(Enumerable.SequenceEqual(uploadBytes, downloadBytes));
        }

        [TestMethod]
        public void UploadDownloadBinaryToBlobStream1()
        {
            var uploadStream = new MemoryStream(TestBytes);
            AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, uploadStream);
            var downloadStream = new MemoryStream();
            AmazonS3Storage.DownloadBlobAsStream(TestBlobAddress, downloadStream);
            Assert.IsTrue(Enumerable.SequenceEqual(uploadStream.ToArray(), downloadStream.ToArray()));
        }

        [TestMethod]
        public void UploadDownloadBinaryToBlobStream2()
        {
            var uploadStream = new MemoryStream(TestBytes);
            AmazonS3Storage.UploadBinaryToBlob(TestBucketName, TestBlobName, uploadStream);
            var downloadStream = new MemoryStream();
            AmazonS3Storage.DownloadBlobAsStream(TestBucketName, TestBlobName, downloadStream);
            Assert.IsTrue(Enumerable.SequenceEqual(uploadStream.ToArray(), downloadStream.ToArray()));
        }

        [TestMethod]
        public void UploadDownloadBinaryToBlobFile1()
        {
            var uploadFileName = Path.GetTempFileName();
            var downloadFileName = Path.GetTempFileName();
            try
            {
                var uploadBytes = TestBytes;
                File.WriteAllBytes(uploadFileName, uploadBytes);
                AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, uploadFileName);
                AmazonS3Storage.DownloadBlobToFile(TestBlobAddress, downloadFileName);
                var downloadBytes = File.ReadAllBytes(downloadFileName);
                Assert.IsTrue(Enumerable.SequenceEqual(uploadBytes, downloadBytes));
            }
            finally
            {
                File.Delete(uploadFileName);
                File.Delete(downloadFileName);
            }
        }

        [TestMethod]
        public void UploadDownloadBinaryToBlobFile2()
        {
            var uploadFileName = Path.GetTempFileName();
            var downloadFileName = Path.GetTempFileName();
            try
            {
                var uploadBytes = TestBytes;
                File.WriteAllBytes(uploadFileName, uploadBytes);
                AmazonS3Storage.UploadBinaryToBlob(TestBucketName, TestBlobName, uploadFileName);
                AmazonS3Storage.DownloadBlobToFile(TestBucketName, TestBlobName, downloadFileName);
                var downloadBytes = File.ReadAllBytes(downloadFileName);
                Assert.IsTrue(Enumerable.SequenceEqual(uploadBytes, downloadBytes));
            }
            finally
            {
                File.Delete(uploadFileName);
                File.Delete(downloadFileName);
            }
        }

        [TestMethod]
        public void DownloadBinaryAllMatch()
        {
            var downloadFileName = Path.GetTempFileName();
            try
            {
                var uploadBytes = TestBytes;
                AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, uploadBytes);
                var downloadBytesArray = AmazonS3Storage.DownloadBlobAsByteArray(TestBlobAddress);
                var downloadStream = new MemoryStream();
                AmazonS3Storage.DownloadBlobAsStream(TestBlobAddress, downloadStream);
                var downloadBytesStream = downloadStream.ToArray();
                AmazonS3Storage.DownloadBlobToFile(TestBlobAddress, downloadFileName);
                var downloadBytesFile = File.ReadAllBytes(downloadFileName);
                Assert.IsTrue(Enumerable.SequenceEqual(downloadBytesArray, downloadBytesStream));
                Assert.IsTrue(Enumerable.SequenceEqual(downloadBytesArray, downloadBytesFile));
            }
            finally
            {
                File.Delete(downloadFileName);
            }
        }

        [TestMethod]
        public void UploadDownloadBinaryToBlobBytesMixed()
        {
            var uploadBytes = TestBytes;
            AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, uploadBytes);
            var downloadBytes = AmazonS3Storage.DownloadBlobAsByteArray(TestBucketName, TestBlobName);
            Assert.IsTrue(Enumerable.SequenceEqual(uploadBytes, downloadBytes));
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void InvalidBlobAddress()
        {
            AmazonS3Storage.DownloadBlobAsByteArray("Invalid");
        }

        [TestMethod]
        [ExpectedException(typeof(WebException))]
        public void InvalidAccessKeyIdList()
        {
            var savedAccessKeyId = AmazonS3Storage.AccessKeyId;
            try
            {
                AmazonS3Storage.AccessKeyId = "INVALID";
                AmazonS3Storage.ListBlobs(TestBucketName);
            }
            finally
            {
                AmazonS3Storage.AccessKeyId = savedAccessKeyId;
            }
        }

        [TestMethod]
        [ExpectedException(typeof(WebException))]
        public void InvalidAccessKeyIdUpload()
        {
            var savedAccessKeyId = AmazonS3Storage.AccessKeyId;
            try
            {
                AmazonS3Storage.AccessKeyId = "INVALID";
                AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, TestBytes);
            }
            finally
            {
                AmazonS3Storage.AccessKeyId = savedAccessKeyId;
            }
        }

        [TestMethod]
        [ExpectedException(typeof(WebException))]
        public void InvalidSecretAccessKeyDownload()
        {
            AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, TestBytes);
            var savedSecretAccessKey = AmazonS3Storage.SecretAccessKey;
            try
            {
                AmazonS3Storage.SecretAccessKey = "INVALID";
                AmazonS3Storage.DownloadBlobAsByteArray(TestBlobAddress);
            }
            finally
            {
                AmazonS3Storage.SecretAccessKey = savedSecretAccessKey;
            }
        }

        [TestMethod]
        [ExpectedException(typeof(WebException))]
        public void InvalidBucketDelete()
        {
            AmazonS3Storage.DeleteBlobContainer("INVALID");
        }

        [TestMethod]
        [ExpectedException(typeof(WebException))]
        public void InvalidDeleteFromBucket()
        {
            AmazonS3Storage.UploadBinaryToBlob(TestBlobAddress, TestBytes);
            AmazonS3Storage.DeleteBlob("INVALID", TestBlobName);
        }
    }
}
