﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using BlobStoreApi;

namespace Delay.Web.Helpers
{
    /// <summary>
    /// Class that provides list/get/set/delete access to Amazon S3 blobs.
    /// </summary>
    /// <remarks>
    /// API based on WindowsAzureStorage from Microsoft.Web.Helpers.WindowsAzureStorage.dll.
    /// </remarks>
    public static class AmazonS3Storage
    {
        // Account information

        /// <summary>
        /// Gets or sets the Access Key ID for the account.
        /// </summary>
        public static string AccessKeyId { get; set; }

        /// <summary>
        /// Gets or sets the Secret Access Key for the account.
        /// </summary>
        public static string SecretAccessKey { get; set; }


        // Blob containers (buckets)

        /// <summary>
        /// Lists the blob containers for an account.
        /// </summary>
        /// <returns>List of container names.</returns>
        public static IList<string> ListBlobContainers()
        {
            var client = GetClient(null);
            return client.GetContainers().ToList();
        }

        /// <summary>
        /// Creates a blob container.
        /// </summary>
        /// <param name="containerName">Container name.</param>
        public static void CreateBlobContainer(string containerName)
        {
            if (null == containerName)
            {
                throw new ArgumentNullException("containerName");
            }
            var client = GetClient(null);
            client.CreateContainer(containerName);
        }

        /// <summary>
        /// Deletes an empty blob container.
        /// </summary>
        /// <param name="containerName">Container name.</param>
        public static void DeleteBlobContainer(string containerName)
        {
            if (null == containerName)
            {
                throw new ArgumentNullException("containerName");
            }
            var client = GetClient(null);
            client.DeleteContainer(containerName);
        }


        // Blobs

        /// <summary>
        /// Lists the blobs in a bucket.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <returns>List of blob addresses.</returns>
        public static IList<string> ListBlobs(string bucketName)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            var client = GetClient(bucketName);
            return client.GetBlobInfos().Select(bi => GetBucketAddress(bucketName, bi.Name)).ToList();
        }

        /// <summary>
        /// Deletes a blob.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        public static void DeleteBlob(string blobAddress)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            DeleteBlob(bucketAndName.Bucket, bucketAndName.Name);
        }

        /// <summary>
        /// Deletes a blob.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        public static void DeleteBlob(string bucketName, string blobName)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            var client = GetClient(bucketName);
            client.DeleteBlob(new BlobInfo(blobName));
        }

        /// <summary>
        /// Creates (or updates) the contents of a blob from a Stream.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <param name="stream">Stream with blob content.</param>
        public static void UploadBinaryToBlob(string blobAddress, Stream stream)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            if (null == stream)
            {
                throw new ArgumentNullException("stream");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            UploadBinaryToBlob(bucketAndName.Bucket, bucketAndName.Name, stream);
        }

        /// <summary>
        /// Creates (or updates) the contents of a blob from a Stream.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <param name="stream">Stream with blob content.</param>
        public static void UploadBinaryToBlob(string bucketName, string blobName, Stream stream)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            if (null == stream)
            {
                throw new ArgumentNullException("stream");
            }
            var client = GetClient(bucketName);
            client.PutBlob(
                new BlobInfo(blobName),
                stream.Length,
                s =>
                {
                    stream.CopyTo(s);
                });
        }

        /// <summary>
        /// Fetches the contents of a blob into a Stream.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <param name="stream">Stream for blob content.</param>
        public static void DownloadBlobAsStream(string blobAddress, Stream stream)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            if (null == stream)
            {
                throw new ArgumentNullException("stream");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            DownloadBlobAsStream(bucketAndName.Bucket, bucketAndName.Name, stream);
        }

        /// <summary>
        /// Fetches the contents of a blob into a Stream.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <param name="stream">Stream for blob content.</param>
        public static void DownloadBlobAsStream(string bucketName, string blobName, Stream stream)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            if (null == stream)
            {
                throw new ArgumentNullException("stream");
            }
            var client = GetClient(bucketName);
            client.GetBlob(
                new BlobInfo(blobName),
                s =>
                {
                    s.CopyTo(stream);
                });
        }

        /// <summary>
        /// Creates (or updates) the contents of a blob from a byte array.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <param name="content">Byte array with blob content.</param>
        public static void UploadBinaryToBlob(string blobAddress, byte[] content)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            if (null == content)
            {
                throw new ArgumentNullException("content");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            UploadBinaryToBlob(bucketAndName.Bucket, bucketAndName.Name, content);
        }

        /// <summary>
        /// Creates (or updates) the contents of a blob from a byte array.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <param name="content">Byte array with blob content.</param>
        public static void UploadBinaryToBlob(string bucketName, string blobName, byte[] content)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            if (null == content)
            {
                throw new ArgumentNullException("content");
            }
            using (var stream = new MemoryStream(content))
            {
                UploadBinaryToBlob(bucketName, blobName, stream);
            }
        }

        /// <summary>
        /// Fetches the contents of a blob into a byte array.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <returns>Byte array with blob content.</returns>
        public static byte[] DownloadBlobAsByteArray(string blobAddress)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            return DownloadBlobAsByteArray(bucketAndName.Bucket, bucketAndName.Name);
        }

        /// <summary>
        /// Fetches the contents of a blob into a byte array.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <returns>Byte array with blob content.</returns>
        public static byte[] DownloadBlobAsByteArray(string bucketName, string blobName)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            using (var stream = new MemoryStream())
            {
                DownloadBlobAsStream(bucketName, blobName, stream);
                return stream.ToArray();
            }
        }

        /// <summary>
        /// Creates (or updates) the contents of a blob from a file.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <param name="fileName">File name with blob content.</param>
        public static void UploadBinaryToBlob(string blobAddress, string fileName)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            if (null == fileName)
            {
                throw new ArgumentNullException("fileName");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            UploadBinaryToBlob(bucketAndName.Bucket, bucketAndName.Name, fileName);
        }

        /// <summary>
        /// Creates (or updates) the contents of a blob from a file.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <param name="fileName">File name with blob content.</param>
        public static void UploadBinaryToBlob(string bucketName, string blobName, string fileName)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            if (null == fileName)
            {
                throw new ArgumentNullException("fileName");
            }
            using (var stream = File.OpenRead(fileName))
            {
                UploadBinaryToBlob(bucketName, blobName, stream);
            }
        }

        /// <summary>
        /// Fetches the contents of a blob into a file.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <param name="fileName">File for blob content.</param>
        public static void DownloadBlobToFile(string blobAddress, string fileName)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            if (null == fileName)
            {
                throw new ArgumentNullException("fileName");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            DownloadBlobToFile(bucketAndName.Bucket, bucketAndName.Name, fileName);
        }

        /// <summary>
        /// Fetches the contents of a blob into a file.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <param name="fileName">File for blob content.</param>
        public static void DownloadBlobToFile(string bucketName, string blobName, string fileName)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            if (null == fileName)
            {
                throw new ArgumentNullException("fileName");
            }
            using (var stream = File.Create(fileName))
            {
                DownloadBlobAsStream(bucketName, blobName, stream);
            }
        }

        /// <summary>
        /// Creates (or updates) the (UTF8-encoded) contents of a blob from a string.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <param name="content">String with blob content.</param>
        public static void UploadTextToBlob(string blobAddress, string content)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            if (null == content)
            {
                throw new ArgumentNullException("content");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            UploadTextToBlob(bucketAndName.Bucket, bucketAndName.Name, content);
        }

        /// <summary>
        /// Creates (or updates) the (UTF8-encoded) contents of a blob from a string.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <param name="content">String with blob content.</param>
        public static void UploadTextToBlob(string bucketName, string blobName, string content)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            if (null == content)
            {
                throw new ArgumentNullException("content");
            }
            UploadBinaryToBlob(bucketName, blobName, Encoding.UTF8.GetBytes(content));
        }

        /// <summary>
        /// Fetches the (UTF8-encoded) contents of a blob into a string.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <returns>String with blob content.</returns>
        public static string DownloadBlobAsText(string blobAddress)
        {
            if (null == blobAddress)
            {
                throw new ArgumentNullException("blobAddress");
            }
            var bucketAndName = GetBucketAndName(blobAddress);
            return DownloadBlobAsText(bucketAndName.Bucket, bucketAndName.Name);
        }

        /// <summary>
        /// Fetches the (UTF8-encoded) contents of a blob into a string.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <param name="blobName">Blob name.</param>
        /// <returns>String with blob content.</returns>
        public static string DownloadBlobAsText(string bucketName, string blobName)
        {
            if (null == bucketName)
            {
                throw new ArgumentNullException("bucketName");
            }
            if (null == blobName)
            {
                throw new ArgumentNullException("blobName");
            }
            var bytes = DownloadBlobAsByteArray(bucketName, blobName);
            return Encoding.UTF8.GetString(bytes);
        }

        /// <summary>
        /// Gets a S3BlobStoreClient instance for a specified bucket name.
        /// </summary>
        /// <param name="bucketName">Bucket name.</param>
        /// <returns>S3BlobStoreClient instance.</returns>
        private static BlobStoreClient GetClient(string bucketName)
        {
            return new S3BlobStoreClient(AccessKeyId, SecretAccessKey, bucketName);
        }

        /// <summary>
        /// Gets a BucketAndName instance for the specified blob address.
        /// </summary>
        /// <param name="blobAddress">Blob address.</param>
        /// <returns>BucketAndName instance.</returns>
        private static BucketAndName GetBucketAndName(string blobAddress)
        {
            var blobAddressParts = blobAddress.Split('/');
            if (2 != blobAddressParts.Length)
            {
                throw new ArgumentException("Malformed blob address (should be \"bucket/name\").", "blobAddress");
            }
            return new BucketAndName(blobAddressParts[0], blobAddressParts[1]);
        }

        /// <summary>
        /// Gets a bucket address for the specified bucket and name.
        /// </summary>
        /// <param name="bucket">Containing bucket.</param>
        /// <param name="name">Blob name.</param>
        /// <returns>Blob address.</returns>
        private static string GetBucketAddress(string bucket, string name)
        {
            return bucket + "/" + name;
        }

        /// <summary>
        /// Class that represents a blob address (bucket and name).
        /// </summary>
        private class BucketAndName
        {
            /// <summary>
            /// Gets the bucket.
            /// </summary>
            public string Bucket { get; private set; }

            /// <summary>
            /// Gets the name.
            /// </summary>
            public string Name { get; private set; }

            /// <summary>
            /// Initializes a new instance of the BucketAndName class.
            /// </summary>
            /// <param name="bucket">Blob bucket.</param>
            /// <param name="name">Blob name.</param>
            public BucketAndName(string bucket, string name)
            {
                Bucket = bucket;
                Name = name;
            }
        }
    }
}
