﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.IO.IsolatedStorage;
using System.Linq;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.Xml.Linq;
#if SILVERLIGHT
using System.Net.Browser;
#else
using System.Threading;
#endif

namespace BlobStoreApi
{
    // Class that represents a simple blob
#if BLOBSTOREAPI_PUBLIC
    public
#endif
    class BlobInfo
    {
        public string Name { get; private set; }
#if !BLOBSTOREAPI_NO_URIS
        public Uri Uri { get; private set; }
        public Uri FriendlyUri { get; private set; }
#endif

        public BlobInfo(string name)
        {
            Name = name;
        }

#if !BLOBSTOREAPI_NO_URIS
        public BlobInfo(string name, Uri baseUri) : this(name)
        {
            Name = name;
            Uri = new Uri(baseUri, Name);
            FriendlyUri = Uri;
        }

        public BlobInfo(string name, Uri baseUri, Uri friendlyBaseUri)
            : this(name, baseUri)
        {
            FriendlyUri = new Uri(friendlyBaseUri, Name);
        }
#endif
    }

    // Base class for a blob store client
#if BLOBSTOREAPI_PUBLIC
    public
#endif
    abstract class BlobStoreClient
    {
        public abstract void GetContainers(Action<IEnumerable<string>> getContainersCompleted, Action<Exception> getContainersFailed);
        public abstract void CreateContainer(string containerName, Action createContainerCompleted, Action<Exception> createContainerFailed);
        public abstract void DeleteContainer(string containerName, Action deleteContainerCompleted, Action<Exception> deleteContainerFailed);
        public abstract void GetBlobInfos(Action<IEnumerable<BlobInfo>> getBlobInfosCompleted, Action<Exception> getBlobInfosFailed);
        public abstract void GetBlob(BlobInfo blobInfo, Action<Stream> getAction, Action getBlobCompleted, Action<Exception> getBlobFailed);
        public abstract void PutBlob(BlobInfo blobInfo, long length, Action<Stream> putAction, Action putBlobCompleted, Action<Exception> putBlobFailed);
        public abstract void DeleteBlob(BlobInfo blobInfo, Action deleteBlobCompleted, Action<Exception> deleteBlobFailed);

        // Synchronous wrappers
#if !SILVERLIGHT
        public IEnumerable<string> GetContainers()
        {
            IEnumerable<string> containers = null;
            var waitHandle = new AutoResetEvent(false);
            Exception exception = null;
            GetContainers(
                (c) =>
                {
                    containers = c;
                    waitHandle.Set();
                },
                (e) =>
                {
                    exception = e;
                    waitHandle.Set();
                });
            waitHandle.WaitOne();
            if (null != exception)
            {
                throw exception;
            }
            return containers;
        }

        public void CreateContainer(string containerName)
        {
            var waitHandle = new AutoResetEvent(false);
            Exception exception = null;
            CreateContainer(containerName, () => waitHandle.Set(), (e) => { exception = e; waitHandle.Set(); });
            waitHandle.WaitOne();
            if (null != exception)
            {
                throw exception;
            }
        }

        public void DeleteContainer(string containerName)
        {
            var waitHandle = new AutoResetEvent(false);
            Exception exception = null;
            DeleteContainer(containerName, () => waitHandle.Set(), (e) => { exception = e; waitHandle.Set(); });
            waitHandle.WaitOne();
            if (null != exception)
            {
                throw exception;
            }
        }

        public IEnumerable<BlobInfo> GetBlobInfos()
        {
            IEnumerable<BlobInfo> blobInfos = null;
            var waitHandle = new AutoResetEvent(false);
            Exception exception = null;
            GetBlobInfos(
                (bi) =>
                {
                    blobInfos = bi;
                    waitHandle.Set();
                },
                (e) =>
                {
                    exception = e;
                    waitHandle.Set();
                });
            waitHandle.WaitOne();
            if (null != exception)
            {
                throw exception;
            }
            return blobInfos;
        }

        public void GetBlob(BlobInfo blobInfo, Action<Stream> getAction)
        {
            var waitHandle = new AutoResetEvent(false);
            Exception exception = null;
            GetBlob(blobInfo, getAction, () => waitHandle.Set(), (e) => { exception = e; waitHandle.Set(); });
            waitHandle.WaitOne();
            if (null != exception)
            {
                throw exception;
            }
        }

        public void PutBlob(BlobInfo blobInfo, long length, Action<Stream> putAction)
        {
            var waitHandle = new AutoResetEvent(false);
            Exception exception = null;
            PutBlob(blobInfo, length, putAction, () => waitHandle.Set(), (e) => { exception = e; waitHandle.Set(); });
            waitHandle.WaitOne();
            if (null != exception)
            {
                throw exception;
            }
        }

        public void DeleteBlob(BlobInfo blobInfo)
        {
            var waitHandle = new AutoResetEvent(false);
            Exception exception = null;
            DeleteBlob(blobInfo, () => waitHandle.Set(), (e) => { exception = e; waitHandle.Set(); });
            waitHandle.WaitOne();
            if (null != exception)
            {
                throw exception;
            }
        }
#endif
    }

    // Base class for a REST blob store client
#if BLOBSTOREAPI_PUBLIC
    public
#endif
    abstract class RestBlobStoreClient : BlobStoreClient
    {
        protected abstract string CustomHeaderPrefix { get; }
        protected abstract string StringToSignFormat { get; }
        protected abstract string AuthorizationPrefix { get; }
        protected abstract string ResponseXmlns { get; }
        protected virtual bool NeedsBogusIfNoneMatch { get { return false; } }

        protected abstract IEnumerable<NameValuePair> GetStoreHeaders();
        protected abstract HttpWebRequest CreateGetContainersRequest(GetContainersState state);
        protected abstract void ProcessGetContainersResponse(HttpWebResponse response, GetContainersState state);
        protected abstract HttpWebRequest CreateCreateContainerRequest(string containerName);
        protected abstract HttpWebRequest CreateDeleteContainerRequest(string containerName);
        protected abstract HttpWebRequest CreateGetBlobInfosRequest(GetBlobInfosState state);
        protected abstract void ProcessGetBlobInfosResponse(HttpWebResponse response, GetBlobInfosState state);
        protected abstract HttpWebRequest CreateGetBlobRequest(BlobInfo blobInfo);
        protected abstract HttpWebRequest CreatePutBlobRequest(BlobInfo blobInfo, long length);
        protected abstract HttpWebRequest CreateDeleteBlobRequest(BlobInfo blobInfo);

        protected string AccountName { get; private set; }

        private const string Unused = "unused";

        private readonly HMAC _hmac;

#if SILVERLIGHT
        [SuppressMessage("Microsoft.Design", "CA1065:DoNotRaiseExceptionsInUnexpectedLocations", Justification = "Never expected to happen in practice.")]
        static RestBlobStoreClient()
        {
            if (!WebRequest.RegisterPrefix("http://", WebRequestCreator.ClientHttp) || !WebRequest.RegisterPrefix("https://", WebRequestCreator.ClientHttp))
            {
                throw new NotSupportedException("BlobStoreClient APIs require the registration of the ClientHttp stack.");
            }
        }
#endif

        public RestBlobStoreClient(string accountName, HMAC hmac)
        {
            AccountName = accountName;
            _hmac = hmac;
        }

        public override void GetContainers(Action<IEnumerable<string>> getContainersCompleted, Action<Exception> getContainersFailed)
        {
            GetContainers(new GetContainersState(getContainersCompleted, getContainersFailed));
        }

        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Passing all exceptions out to the caller.")]
        protected virtual void GetContainers(GetContainersState getContainersState)
        {
            var request = CreateGetContainersRequest(getContainersState);
            request.BeginGetResponse(
                iAsyncResult =>
                {
                    try
                    {
                        using (var response = (HttpWebResponse)request.EndGetResponse(iAsyncResult))
                        {
                            ProcessGetContainersResponse(response, getContainersState);
                        }
                    }
                    catch (Exception e)
                    {
                        getContainersState.GetFailed(e);
                    }
                },
                null);
        }

        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Passing all exceptions out to the caller.")]
        public override void CreateContainer(string containerName, Action createContainerCompleted, Action<Exception> createContainerFailed)
        {
            var request = CreateCreateContainerRequest(containerName);
            request.BeginGetResponse(
                iAsyncResult =>
                {
                    try
                    {
                        using (var response = (HttpWebResponse)request.EndGetResponse(iAsyncResult))
                        {
                            createContainerCompleted();
                        }
                    }
                    catch (Exception e)
                    {
                        createContainerFailed(e);
                    }
                },
                null);
        }

        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Passing all exceptions out to the caller.")]
        public override void DeleteContainer(string containerName, Action deleteContainerCompleted, Action<Exception> deleteContainerFailed)
        {
            var request = CreateDeleteContainerRequest(containerName);
            request.BeginGetResponse(
                iAsyncResult =>
                {
                    try
                    {
                        using (var response = (HttpWebResponse)request.EndGetResponse(iAsyncResult))
                        {
                            deleteContainerCompleted();
                        }
                    }
                    catch (Exception e)
                    {
                        deleteContainerFailed(e);
                    }
                },
                null);
        }

        public override void GetBlobInfos(Action<IEnumerable<BlobInfo>> getBlobInfosCompleted, Action<Exception> getBlobInfosFailed)
        {
            GetBlobInfos(new GetBlobInfosState(getBlobInfosCompleted, getBlobInfosFailed));
        }

        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Passing all exceptions out to the caller.")]
        protected virtual void GetBlobInfos(GetBlobInfosState getBlobInfosState)
        {
            var request = CreateGetBlobInfosRequest(getBlobInfosState);
            request.BeginGetResponse(
                iAsyncResult =>
                {
                    try
                    {
                        using (var response = (HttpWebResponse)request.EndGetResponse(iAsyncResult))
                        {
                            ProcessGetBlobInfosResponse(response, getBlobInfosState);
                        }
                    }
                    catch (Exception e)
                    {
                        getBlobInfosState.GetFailed(e);
                    }
                },
                null);
        }

        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Passing all exceptions out to the caller.")]
        public override void GetBlob(BlobInfo blobInfo, Action<Stream> getAction, Action getBlobCompleted, Action<Exception> getBlobFailed)
        {
            var request = CreateGetBlobRequest(blobInfo);
#if SILVERLIGHT
            // Do NOT want the entire response stream to be buffered in memory
            request.AllowReadStreamBuffering = false;
#endif
            request.BeginGetResponse(
                iAsyncResult =>
                {
                    try
                    {
                        using (var response = (HttpWebResponse)request.EndGetResponse(iAsyncResult))
                        {
                            using (var stream = response.GetResponseStream())
                            {
                                getAction(stream);
                            }
                            getBlobCompleted();
                        }
                    }
                    catch (Exception e)
                    {
                        getBlobFailed(e);
                    }
                },
                null);
        }

        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Passing all exceptions out to the caller.")]
        public override void PutBlob(BlobInfo blobInfo, long length, Action<Stream> putAction, Action putBlobCompleted, Action<Exception> putBlobFailed)
        {
            var request = CreatePutBlobRequest(blobInfo, length);
#if !WINDOWS_PHONE
            // Do NOT want the entire request stream to be buffered in memory
            request.AllowWriteStreamBuffering = false;
#endif
            request.BeginGetRequestStream(
                iAsyncResultRequest =>
                {
                    try
                    {
                        using (var stream = request.EndGetRequestStream(iAsyncResultRequest))
                        {
                            putAction(stream);
                        }
                        request.BeginGetResponse(
                            iAsyncResultResponse =>
                            {
                                try
                                {
                                    using (request.EndGetResponse(iAsyncResultResponse)) { }
                                    putBlobCompleted();
                                }
                                catch (Exception e)
                                {
                                    putBlobFailed(e);
                                }
                            },
                            null);
                    }
                    catch (Exception e)
                    {
                        putBlobFailed(e);
                    }
                },
                null);
        }

        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Passing all exceptions out to the caller.")]
        public override void DeleteBlob(BlobInfo blobInfo, Action deleteBlobCompleted, Action<Exception> deleteBlobFailed)
        {
            var request = CreateDeleteBlobRequest(blobInfo);
            request.BeginGetResponse(
                iAsyncResult =>
                {
                    try
                    {
                        using (var response = (HttpWebResponse)request.EndGetResponse(iAsyncResult))
                        {
                            deleteBlobCompleted();
                        }
                    }
                    catch (Exception e)
                    {
                        deleteBlobFailed(e);
                    }
                },
                null);
        }

        protected HttpWebRequest CreateRequest(string baseUriString, string canonicalizedResourceFormat, string method, string path, long? length, IEnumerable<NameValuePair> headers, IEnumerable<NameValuePair> parameters)
        {
            // Sanitize parameters
            if (null == headers)
            {
                headers = Enumerable.Empty<NameValuePair>();
            }
            if (null == parameters)
            {
                parameters = Enumerable.Empty<NameValuePair>();
            }

            // Create request object
            var requestUri = new Uri(
                baseUriString +
                ((null != path) ? path : "") +
                (parameters.Any() ?
                    "?" + string.Join("&", parameters.Select(p => p.Name + "=" + p.Value).ToArray()) :
                    ""));
            var request = (HttpWebRequest)WebRequest.Create(requestUri);
            request.Method = method.ToUpper(CultureInfo.InvariantCulture);
#if !SILVERLIGHT
            request.Timeout = int.MaxValue;
#endif
#if !WINDOWS_PHONE
            if (length.HasValue)
            {
                request.ContentLength = length.Value;
            }
#endif
            // The following is necessary to prevent the Cache-Control header from
            // getting sent up and preventing successful download when not behind
            // a proxy.
            request.Headers[HttpRequestHeader.CacheControl] = Unused;
            // The following may be necessary to prevent the If-None-Match and
            // If-Modified-Since headers from getting sent up and preventing
            // successful calls due to an Authorization header mis-match.
            if (NeedsBogusIfNoneMatch)
            {
                request.Headers[HttpRequestHeader.IfNoneMatch] = Unused;
            }

            // Process headers
            var orderedHeaders =
                Enumerable.Concat(GetStoreHeaders(), headers)
                    .OrderBy(p => ToLowerInvariant(p.Name))
                    .ToArray();
            foreach (var header in orderedHeaders)
            {
                request.Headers[header.Name] = header.Value;
            }

            // Canonicalize headers
            var canonicalizedHeaders =
                string.Concat(
                    orderedHeaders
                        .Where(p => ToLowerInvariant(p.Name).StartsWith(CustomHeaderPrefix, StringComparison.Ordinal))
                        .Select(p => ToLowerInvariant(p.Name) + ":" + p.Value + "\n")
                        .ToArray());

            // Canonicalize resource
            var canonicalizedResource =
                string.Format(
                    CultureInfo.InvariantCulture,
                    ToLowerInvariant(canonicalizedResourceFormat),
                    AccountName,
                    requestUri.AbsolutePath,
                    string.Concat(
                        parameters
                            .OrderBy(p => p.Name)
                            .Select(p => "\n" + p.Name + ":" + p.Value)
                            .ToArray()));

            // Create Authorization header
            var stringToSign =
                string.Format(
                    CultureInfo.InvariantCulture,
                    StringToSignFormat,
                    request.Method,
                    null,
                    null,
                    (length.HasValue ? length.Value.ToString(CultureInfo.InvariantCulture) : null),
                    null,
                    null,
                    null,
                    null,
                    null,
                    NeedsBogusIfNoneMatch ? Unused : null,
                    null,
                    null,
                    canonicalizedHeaders,
                    canonicalizedResource);
            var signature = Convert.ToBase64String(_hmac.ComputeHash(Encoding.UTF8.GetBytes(stringToSign)));
            var authorization = AuthorizationPrefix + " " + AccountName + ":" + signature;
            request.Headers[HttpRequestHeader.Authorization] = authorization;

            // Return prepared request object
            return request;
        }

        [SuppressMessage("Microsoft.Globalization", "CA1308:NormalizeStringsToUppercase", Justification = "Lower-case required by protocol.")]
        protected static string ToLowerInvariant(string s)
        {
            return s.ToLowerInvariant();
        }

        protected class GetState<T>
        {
            public Action<IEnumerable<T>> GetCompleted { get; private set; }
            public Action<Exception> GetFailed { get; private set; }
            public List<T> Items { get; private set; }
            public string Marker { get; set; }
            public GetState(Action<IEnumerable<T>> getCompleted, Action<Exception> getFailed)
            {
                GetCompleted = getCompleted;
                GetFailed = getFailed;
                Items = new List<T>();
            }
        }

        protected class GetBlobInfosState : GetState<BlobInfo>
        {
            public GetBlobInfosState(Action<IEnumerable<BlobInfo>> getBlobInfosCompleted, Action<Exception> getBlobInfosFailed)
                : base(getBlobInfosCompleted, getBlobInfosFailed)
            {
            }
        }

        protected class GetContainersState : GetState<string>
        {
            public GetContainersState(Action<IEnumerable<string>> getContainersCompleted, Action<Exception> getContainersFailed)
                : base(getContainersCompleted, getContainersFailed)
            {
            }
        }

        protected class NameValuePair
        {
            public string Name { get; private set; }
            public string Value { get; private set; }
            public NameValuePair(string name, string value)
            {
                Name = name;
                Value = value;
            }
        }
    }

#if !BLOBSTOREAPI_NO_AZUREBLOBSTORECLIENT
    // Blob store client for the Windows Azure service
#if BLOBSTOREAPI_PUBLIC
    public
#endif
    class AzureBlobStoreClient : RestBlobStoreClient
    {
        public const string RootContainerName = "$root";

        private readonly string _containerName;

        [SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope", Justification = "Object is reused by the instance.")]
        public AzureBlobStoreClient(string accountName, string accessKey, string containerName)
            : base(accountName, new HMACSHA256(Convert.FromBase64String(accessKey)))
        {
            _containerName = (null != containerName) ? ToLowerInvariant(containerName) : null;
        }

        protected override string CustomHeaderPrefix { get { return "x-ms-"; } }
        protected override string AuthorizationPrefix { get { return "SharedKey"; } }
        protected override string ResponseXmlns { get { return ""; } }
        protected override bool NeedsBogusIfNoneMatch { get { return true; } }
        protected override string StringToSignFormat
        {
            get
            {
                return
                    /* Request method */ "{0}\n" +
                    /* Content-Encoding */ "{1}\n" +
                    /* Content-Language */ "{2}\n" +
                    /* Content-Length */ "{3}\n" +
                    /* Content-MD5 */ "{4}\n" +
                    /* Content-Type */ "{5}\n" +
                    /* Date */ "{6}\n" +
                    /* If-Modified-Since */ "{7}\n" +
                    /* If-Match */ "{8}\n" +
                    /* If-None-Match */ "{9}\n" +
                    /* If-Unmodified-Since */ "{10}\n" +
                    /* Range */ "{11}\n" +
                    /* Canonicalized headers */ "{12}" +
                    /* Canonicalized resource */ "{13}";
            }
        }

        protected string ContainerRequestUriString { get { return string.Format(CultureInfo.InvariantCulture, "http://{0}.blob.core.windows.net/", AccountName); } }
        protected string BlobRequestUriString { get { return string.Format(CultureInfo.InvariantCulture, "http://{0}.blob.core.windows.net/{1}/", AccountName, _containerName); } }
        protected static string CanonicalizedResourceFormat { get { return "/{0}{1}{2}"; } }

        protected override IEnumerable<NameValuePair> GetStoreHeaders()
        {
            yield return new NameValuePair("x-ms-date", DateTime.UtcNow.ToString("r", CultureInfo.InvariantCulture));
            yield return new NameValuePair("x-ms-version", "2009-09-19");
        }

        protected override HttpWebRequest CreateGetContainersRequest(GetContainersState state)
        {
            var parameters = new List<NameValuePair>();
            parameters.Add(new NameValuePair("comp", "list"));
#if BLOBSTOREAPI_FRAGMENTGETBLOBINFOSFORTESTING
            parameters.Add(new NameValuePair("maxresults", "3"));
#endif
            if (null != state.Marker)
            {
                parameters.Add(new NameValuePair("marker", state.Marker));
            }
            return CreateRequest(ContainerRequestUriString, CanonicalizedResourceFormat, "GET", null, null, null, parameters);
        }

        protected override void ProcessGetContainersResponse(HttpWebResponse response, GetContainersState state)
        {
            ProcessChainedRequest(
                response,
                state,
                "Containers",
                "Container",
                (e) => e.Element(XName.Get("Name", ResponseXmlns)).Value,
                () => GetContainers(state));
        }

        protected override HttpWebRequest CreateCreateContainerRequest(string containerName)
        {
            return CreateRequest(
                ContainerRequestUriString,
                CanonicalizedResourceFormat,
                "PUT",
                ToLowerInvariant(containerName),
                0,
                null,
                new NameValuePair[] { new NameValuePair("restype", "container") });
        }

        protected override HttpWebRequest CreateDeleteContainerRequest(string containerName)
        {
            return CreateRequest(
                ContainerRequestUriString,
                CanonicalizedResourceFormat,
                "DELETE",
                ToLowerInvariant(containerName),
                0,
                null,
                new NameValuePair[] { new NameValuePair("restype", "container") });
        }

        protected override HttpWebRequest CreateGetBlobInfosRequest(GetBlobInfosState state)
        {
            var parameters = new List<NameValuePair>();
            parameters.Add(new NameValuePair("restype", "container"));
            parameters.Add(new NameValuePair("comp", "list"));
#if BLOBSTOREAPI_FRAGMENTGETBLOBINFOSFORTESTING
            parameters.Add(new NameValuePair("maxresults", "3"));
#endif
            if (null != state.Marker)
            {
                parameters.Add(new NameValuePair("marker", state.Marker));
            }
            return CreateRequest(BlobRequestUriString, CanonicalizedResourceFormat, "GET", null, null, null, parameters);
        }

        protected override void ProcessGetBlobInfosResponse(HttpWebResponse response, GetBlobInfosState state)
        {
            ProcessChainedRequest(
                response,
                state,
                "Blobs",
                "Blob",
                (e) => new BlobInfo(
                    e.Element(XName.Get("Name", ResponseXmlns)).Value
#if !BLOBSTOREAPI_NO_URIS
                    , new Uri(BlobRequestUriString)
                    , new Uri(BlobRequestUriString.Replace("$root/", ""))
#endif
                    ),
                () => GetBlobInfos(state));
        }

        protected override HttpWebRequest CreateGetBlobRequest(BlobInfo blobInfo)
        {
            return CreateRequest(BlobRequestUriString, CanonicalizedResourceFormat, "GET", blobInfo.Name, null, null, null);
        }

        protected override HttpWebRequest CreatePutBlobRequest(BlobInfo blobInfo, long length)
        {
            return CreateRequest(
                BlobRequestUriString,
                CanonicalizedResourceFormat,
                "PUT",
                blobInfo.Name,
                length,
                new NameValuePair[]
                {
                    new NameValuePair("x-ms-blob-type", "BlockBlob"),
                },
                null);
        }

        protected override HttpWebRequest CreateDeleteBlobRequest(BlobInfo blobInfo)
        {
            return CreateRequest(BlobRequestUriString, CanonicalizedResourceFormat, "DELETE", blobInfo.Name, 0, null, null);
        }

        protected void ProcessChainedRequest<T>(HttpWebResponse response, GetState<T> state, string collectionElementName, string elementName, Func<XElement, T> getTFromElement, Action fetchNextBatch)
        {
            using (var reader = new StreamReader(response.GetResponseStream()))
            {
                var rootElement = XElement.Load(reader);
                state.Items.AddRange(
                    rootElement
                        .Element(XName.Get(collectionElementName, ResponseXmlns))
                        .Elements(XName.Get(elementName, ResponseXmlns))
                        .Select(e => getTFromElement(e)));
                var nextMarkerElement =
                    rootElement
                        .Element(XName.Get("NextMarker", ResponseXmlns));
                state.Marker = ((null != nextMarkerElement) && !string.IsNullOrEmpty(nextMarkerElement.Value)) ?
                    nextMarkerElement.Value :
                    null;
            }
            if (null != state.Marker)
            {
                fetchNextBatch();
            }
            else
            {
                state.GetCompleted(state.Items);
            }
        }
    }
#endif


#if !BLOBSTOREAPI_NO_S3BLOBSTORECLIENT
    // Blob store client for the Amazon S3 service
#if BLOBSTOREAPI_PUBLIC
    public
#endif
    class S3BlobStoreClient : RestBlobStoreClient
    {
        private readonly string _bucketName;

        [SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope", Justification = "Object is reused by the instance.")]
        public S3BlobStoreClient(string accessKeyId, string secretAccessKey, string bucketName)
            : base(accessKeyId, new HMACSHA1(Encoding.UTF8.GetBytes(secretAccessKey)))
        {
            _bucketName = bucketName;
        }

        protected override string CustomHeaderPrefix { get { return "x-amz-"; } }
        protected override string AuthorizationPrefix { get { return "AWS"; } }
        protected override string ResponseXmlns { get { return "http://s3.amazonaws.com/doc/2006-03-01/"; } }
        protected override string StringToSignFormat
        {
            get
            {
                return
                    /* Request method */ "{0}\n" +
                    /* Content-MD5 */ "{4}\n" +
                    /* Content-Type */ "{5}\n" +
                    /* Date */ "{6}\n" +
                    /* Canonicalized headers */ "{12}" +
                    /* Canonicalized resource */ "{13}";
            }
        }

        protected static string GetContainersRequestUriString { get { return "http://s3.amazonaws.com/"; } }
        protected static string ContainersRequestUriStringFormat { get { return "http://{0}.s3.amazonaws.com/"; } }
        protected string BlobRequestUriString { get { return string.Format(CultureInfo.InvariantCulture, "http://{0}.s3.amazonaws.com/", _bucketName); } }
        protected static string CanonicalizedGetContainerResourceFormat { get { return "{1}"; } }
        protected static string CanonicalizedContainersResourceFormatFormat { get { return "/{0}/"; } }
        protected string CanonicalizedBlobResourceFormat { get { return "/" + ToLowerInvariant(_bucketName) + "{1}"; } }

        protected override IEnumerable<NameValuePair> GetStoreHeaders()
        {
            yield return new NameValuePair("x-amz-date", DateTime.UtcNow.ToString("r", CultureInfo.InvariantCulture));
        }

        protected override HttpWebRequest CreateGetContainersRequest(GetContainersState state)
        {
            return CreateRequest(GetContainersRequestUriString, CanonicalizedGetContainerResourceFormat, "GET", null, null, null, null);
        }

        protected override void ProcessGetContainersResponse(HttpWebResponse response, GetContainersState state)
        {
            using (var reader = new StreamReader(response.GetResponseStream()))
            {
                state.GetCompleted(
                    XElement.Load(reader)
                        .Element(XName.Get("Buckets", ResponseXmlns))
                        .Elements(XName.Get("Bucket", ResponseXmlns))
                        .Select(e => e.Element(XName.Get("Name", ResponseXmlns)).Value));
            }
        }

        protected override HttpWebRequest CreateCreateContainerRequest(string containerName)
        {
            return CreateRequest(
                string.Format(CultureInfo.InvariantCulture, ContainersRequestUriStringFormat, containerName),
                string.Format(CultureInfo.InvariantCulture, CanonicalizedContainersResourceFormatFormat, containerName),
                "PUT",
                null,
                0,
                null,
                null);
        }

        protected override HttpWebRequest CreateDeleteContainerRequest(string containerName)
        {
            return CreateRequest(
                string.Format(CultureInfo.InvariantCulture, ContainersRequestUriStringFormat, containerName),
                string.Format(CultureInfo.InvariantCulture, CanonicalizedContainersResourceFormatFormat, containerName),
                "DELETE",
                null,
                0,
                null,
                null);
        }

        protected override HttpWebRequest CreateGetBlobInfosRequest(GetBlobInfosState state)
        {
            var parameters = new List<NameValuePair>();
#if BLOBSTOREAPI_FRAGMENTGETBLOBINFOSFORTESTING
            parameters.Add(new NameValuePair("max-keys", "3"));
#endif
            if (null != state.Marker)
            {
                parameters.Add(new NameValuePair("marker", state.Marker));
            }
            return CreateRequest(BlobRequestUriString, CanonicalizedBlobResourceFormat, "GET", null, null, null, parameters);
        }

        protected override void ProcessGetBlobInfosResponse(HttpWebResponse response, GetBlobInfosState state)
        {
            using (var reader = new StreamReader(response.GetResponseStream()))
            {
                var rootElement = XElement.Load(reader);
                state.Items.AddRange(
                    rootElement
                        .Elements(XName.Get("Contents", ResponseXmlns))
                        .Select(e => new BlobInfo(
                            e.Element(XName.Get("Key", ResponseXmlns)).Value
#if !BLOBSTOREAPI_NO_URIS
                            , new Uri(BlobRequestUriString)
#endif
                        )));
                var isTruncatedElement =
                    rootElement
                        .Element(XName.Get("IsTruncated", ResponseXmlns));
                state.Marker = ((null != isTruncatedElement) && string.Equals(isTruncatedElement.Value, "true", StringComparison.OrdinalIgnoreCase)) ?
                        state.Items.Last().Name :
                        null;
            }
            if (null != state.Marker)
            {
                // Need to fetch the next batch
                GetBlobInfos(state);
            }
            else
            {
                // Done; return results
                state.GetCompleted(state.Items);
            }
        }

        protected override HttpWebRequest CreateGetBlobRequest(BlobInfo blobInfo)
        {
            return CreateRequest(BlobRequestUriString, CanonicalizedBlobResourceFormat, "GET", blobInfo.Name, null, null, null);
        }

        protected override HttpWebRequest CreatePutBlobRequest(BlobInfo blobInfo, long length)
        {
            return CreateRequest(
                BlobRequestUriString,
                CanonicalizedBlobResourceFormat,
                "PUT",
                blobInfo.Name,
                length,
                new NameValuePair[]
                {
                    new NameValuePair("x-amz-acl", "public-read"),
                },
                null);
        }

        protected override HttpWebRequest CreateDeleteBlobRequest(BlobInfo blobInfo)
        {
            return CreateRequest(BlobRequestUriString, CanonicalizedBlobResourceFormat, "DELETE", blobInfo.Name, 0, null, null);
        }
    }
#endif

#if SILVERLIGHT && !BLOBSTOREAPI_NO_ISOLATEDSTORAGEBLOBSTORECLIENT
    // Blob store client for Silverlight isolated storage
    // (TEST-only; browser can't open links from inside isolated storage)
#if BLOBSTOREAPI_PUBLIC
    public
#endif
    class IsolatedStorageBlobStoreClient : BlobStoreClient
    {
        private IsolatedStorageFile _isolatedStorageFile = IsolatedStorageFile.GetUserStoreForApplication();

        public override void GetContainers(Action<IEnumerable<string>> getContainersCompleted, Action<Exception> getContainersFailed)
        {
            throw new NotImplementedException();
        }

        public override void CreateContainer(string containerName, Action createContainerCompleted, Action<Exception> createContainerFailed)
        {
            throw new NotImplementedException();
        }

        public override void DeleteContainer(string containerName, Action deleteContainerCompleted, Action<Exception> deleteContainerFailed)
        {
            throw new NotImplementedException();
        }

        public override void GetBlobInfos(Action<IEnumerable<BlobInfo>> getBlobInfosCompleted, Action<Exception> getBlobInfosFailed)
        {
            getBlobInfosCompleted(_isolatedStorageFile.GetFileNames()
                .Where(n => 0 != string.CompareOrdinal(n, "__LocalSettings"))
                .Select(n => new BlobInfo(n, new Uri(new Uri("http://localhost/$root/"), n), new Uri("http://localhost/"))));
        }

        public override void GetBlob(BlobInfo blobInfo, Action<Stream> getAction, Action getBlobCompleted, Action<Exception> getBlobFailed)
        {
            using (var stream = _isolatedStorageFile.OpenFile(blobInfo.Name, FileMode.Open))
            {
                getAction(stream);
            }
            getBlobCompleted();
        }

        public override void PutBlob(BlobInfo blobInfo, long length, Action<Stream> putAction, Action putBlobCompleted, Action<Exception> putBlobFailed)
        {
            using (var stream = _isolatedStorageFile.CreateFile(blobInfo.Name))
            {
                putAction(stream);
            }
            putBlobCompleted();
        }

        public override void DeleteBlob(BlobInfo blobInfo, Action deleteBlobCompleted, Action<Exception> deleteBlobFailed)
        {
            _isolatedStorageFile.DeleteFile(blobInfo.Name);
            deleteBlobCompleted();
        }
    }
#endif
}
