﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Mime;
using System.Web;

namespace Delay.Web.Helpers
{
    /// <summary>
    /// Class that creates "data" URIs for embedding linked content inside a web page.
    /// </summary>
    /// <remarks>
    /// See also:
    /// http://en.wikipedia.org/wiki/Data_Uri
    /// http://tools.ietf.org/html/rfc2397
    /// </remarks>
    public static class DataUri
    {
        /// <summary>
        /// Writes a data URI for the specified file (inferring media type from the file extension).
        /// </summary>
        /// <param name="fileName">File to create a data URI for.</param>
        /// <returns>Data URI.</returns>
        public static HtmlString Write(string fileName)
        {
            if (null == fileName)
            {
                throw new ArgumentNullException("fileName");
            }

            // Determine media type
            string mediaType = MediaTypeNames.Application.Octet;
            switch (Path.GetExtension(fileName).ToUpperInvariant())
            {
                case ".GIF":
                    mediaType = MediaTypeNames.Image.Gif;
                    break;
                case ".JPG":
                case ".JPEG":
                    mediaType = MediaTypeNames.Image.Jpeg;
                    break;
                case ".PNG":
                    mediaType = "image/png"; // Not provided by MediaTypeNames
                    break;
                case ".CSS":
                    mediaType = "text/css"; // Not provided by MediaTypeNames
                    break;
                case ".JS":
                    mediaType = "text/javascript"; // Not provided by MediaTypeNames
                    break;
                case ".HTM":
                case ".HTML":
                    mediaType = MediaTypeNames.Text.Html;
                    break;
                case ".XML":
                    mediaType = MediaTypeNames.Text.Xml;
                    break;
                case ".TXT":
                    mediaType = MediaTypeNames.Text.Plain;
                    break;
            }

            // Write result
            return Write(fileName, mediaType);
        }

        /// <summary>
        /// Writes a data URI for the specified file and media type.
        /// </summary>
        /// <param name="fileName">File to create a data URI for.</param>
        /// <param name="mediaType">Media type for the data URI.</param>
        /// <returns>Data URI.</returns>
        public static HtmlString Write(string fileName, string mediaType)
        {
            if (null == fileName)
            {
                throw new ArgumentNullException("fileName");
            }
            if (null == mediaType)
            {
                throw new ArgumentNullException("mediaType");
            }

            // Write result
            return Write(File.ReadAllBytes(fileName), mediaType);
        }

        /// <summary>
        /// Writes a data URI for the specified Stream and media type.
        /// </summary>
        /// <param name="data">Stream to create a data URI for.</param>
        /// <param name="mediaType">Media type for the data URI.</param>
        /// <returns>Data URI.</returns>
        public static HtmlString Write(Stream data, string mediaType)
        {
            if (null == data)
            {
                throw new ArgumentNullException("data");
            }
            if (null == mediaType)
            {
                throw new ArgumentNullException("mediaType");
            }

            // Write result
            return Write(ConvertStreamToSequence(data), mediaType);
        }

        /// <summary>
        /// Writes a data URI for the specified sequence of bytes and media type.
        /// </summary>
        /// <param name="data">Sequence of bytes to create a data URI for.</param>
        /// <param name="mediaType">Media type for the data URI.</param>
        /// <returns>Data URI.</returns>
        public static HtmlString Write(IEnumerable<byte> data, string mediaType)
        {
            if (null == data)
            {
                throw new ArgumentNullException("data");
            }
            if (null == mediaType)
            {
                throw new ArgumentNullException("mediaType");
            }

            // Write result
            return Write(data.ToArray(), mediaType);
        }

        /// <summary>
        /// Writes a data URI for the specified array of bytes and media type.
        /// </summary>
        /// <param name="data">Array of bytes to create a data URI for.</param>
        /// <param name="mediaType">Media type for the data URI.</param>
        /// <returns>Data URI.</returns>
        public static HtmlString Write(byte[] data, string mediaType)
        {
            if (null == data)
            {
                throw new ArgumentNullException("data");
            }
            if (null == mediaType)
            {
                throw new ArgumentNullException("mediaType");
            }

            // Write result
            return new HtmlString(string.Concat("data:", mediaType, ";base64,", Convert.ToBase64String(data)));
        }

        /// <summary>
        /// Converts the specified Stream to a sequence of bytes.
        /// </summary>
        /// <param name="stream">Stream to convert.</param>
        /// <returns>Sequence of bytes.</returns>
        private static IEnumerable<byte> ConvertStreamToSequence(Stream stream)
        {
            int i;
            while (-1 != (i = stream.ReadByte()))
            {
                yield return (byte)i;
            }
        }
    }
}
